// Algorithme Réseau de Capteurs SRF08 
// PFE : Module d'aide au déplacement pour fauteuil roulant électrique
// Auteurs : Geoffrey ROSE & Marjorie Tixier
// IMA5 Polytech'Lille
// 
// Dans ce code, on utilise 7 capteurs SRF08 en réseau afin de détecter des obstacles dans 4 directions (Avant, arrière, droite et gauche)
// En fonction des informations analogiques lues (qui reflétent la position du joystick) et des distances d'obstacles, on calcule une tension PWM
// permettant de remplacer l'information du joystick si l'obstacle est considéré comme trop proche. Cette tension PWM sera ensuite transformé en tension continue
// via un filtre RC et un AOP.
//


#include <Wire.h>
#include <SonarSRF08.h>

//Adresses des capteurs SRF08
#define ONE_ADDRESS 0x70
#define TWO_ADDRESS 0x71
#define THREE_ADDRESS 0x72
#define FOUR_ADDRESS 0x73
#define FIVE_ADDRESS 0x74
#define SIX_ADDRESS 0x75
#define SEVEN_ADDRESS 0x76

//Déclaration des objets SonarSRF08
SonarSRF08 SonarOne;
SonarSRF08 SonarTwo;
SonarSRF08 SonarThree;
SonarSRF08 SonarFour;
SonarSRF08 SonarFive;
SonarSRF08 SonarSix;
SonarSRF08 SonarSeven;

// Setup Analogue Gain
#define GAIN_REGISTER 0x09
// Setup Range Location
#define LOCATION_REGISTER 0x8C

char unit = 'c'; // 'i' for inches, 'c' for centimeters, 'm' for micro-seconds

//Déclaration des différentes distances (en cm)
int Av_Slow=100;
int Av_Stop=45;
int Ar_Slow=100;
int Ar_Stop=45;
int Co_Slow=35;
int Co_Stop=15;

//Déclaration des différents ports PWM
int PWM_AA=5;
int PWM_DG=6;

//Déclaration des différents 
int AN_AA=0;
int AN_DG=1;


///////////////////////////////////////////////////////////////////////////////////////////////////
// Setup : Connexion aux capteurs, test de reconnaissance et configuration de la range de 2 mètres.
///////////////////////////////////////////////////////////////////////////////////////////////////
void setup() {
  
    Serial.begin(9600);
    
    SonarOne.connect(ONE_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    SonarTwo.connect(TWO_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    SonarThree.connect(THREE_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    SonarFour.connect(FOUR_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    SonarFive.connect(FIVE_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    SonarSix.connect(SIX_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    SonarSeven.connect(SEVEN_ADDRESS, GAIN_REGISTER, LOCATION_REGISTER);
    
    isConnected("SRF08 n1 : Avant Gauche", SonarOne.getSoft());
    isConnected("SRF08 n2 : Avant Droite", SonarTwo.getSoft());
    isConnected("SRF08 n3 : Droit Avant", SonarThree.getSoft());
    isConnected("SRF08 n4 : Arriere", SonarFour.getSoft());
    isConnected("SRF08 n5 : Droit Arriere", SonarFive.getSoft());
    isConnected("SRF08 n6 : Gauche Arriere", SonarSix.getSoft());
    isConnected("SRF08 n7 : Gauche Avant", SonarSeven.getSoft());

    int RangeRegister = 0x2E;
    
    setupSensor(SonarOne,RangeRegister);
    setupSensor(SonarTwo,RangeRegister);
    setupSensor(SonarThree,RangeRegister);
    setupSensor(SonarFour,RangeRegister);
    setupSensor(SonarFive,RangeRegister);
    setupSensor(SonarSix,RangeRegister);
    setupSensor(SonarSeven, RangeRegister);
    
    pinMode(PWM_AA, OUTPUT);
    pinMode(PWM_DG, OUTPUT);
    
    setPwmFrequency(PWM_AA,1);
    setPwmFrequency(PWM_DG,1);
    
}

void loop() {
    
    // Initialisation des stockages de valeurs
    float SensorOne_Value = 0;
    float SensorTwo_Value = 0;
    float SensorThree_Value = 0;
    float SensorFour_Value = 0;
    float SensorFive_Value = 0;
    float SensorSix_Value = 0;
    float SensorSeven_Value = 0;
    int AN_AA_Val_Octet = 0;                 // Entre 0 et 770
    float AN_AA_Val_Volts = 0;               // Entre 1,1 et 3,9V
    int AN_DG_Val_Octet = 0;                 // Entre  0 et 770
    float AN_DG_Val_Volts = 0;               // Entre 1,1 et 3,9V
    float PWM_AA_Val_Volts = 0;              // Entre 1,1 et 3,9V
    int PWM_AA_Val_Octet = 0;                // Entre 52 et 184
    float PWM_DG_Val_Volts = 0;              // Entre 1,1 et 3,9V
    int PWM_DG_Val_Octet = 0;                // Entre 52 et 184
  
    // Récupération des valeurs en cm
    SensorOne_Value = SonarOne.getRange(unit);
    delay(50);
    SensorTwo_Value = SonarTwo.getRange(unit);
    delay(50);
    SensorThree_Value = SonarThree.getRange(unit);
    delay(50);
    SensorFour_Value = SonarFour.getRange(unit);
    delay(50);
    SensorFive_Value = SonarFive.getRange(unit);
    delay(50);
    SensorSix_Value = SonarSix.getRange(unit);
    delay(50);
    SensorSeven_Value = SonarSeven.getRange(unit);
    delay(50);
    
    //Variables de simulation : On force la valeur des capteurs pour simuler un environnement vide.
    /*SensorOne_Value=150;
    SensorTwo_Value=150;
    SensorThree_Value=150;
    SensorFour_Value=150;
    SensorFive_Value=150;
    SensorSix_Value=150;
    SensorSeven_Value=150;*/
    

    //Lecture des ports analogiques
    AN_AA_Val_Octet=analogRead(AN_AA); // 0 a 1023
    AN_AA_Val_Volts=ConvertAnalogToVolts(AN_AA_Val_Octet); // 1,1 et 3,9    
    
    AN_DG_Val_Octet=analogRead(AN_DG); // 0 a 1023
    AN_DG_Val_Volts=ConvertAnalogToVolts(AN_DG_Val_Octet); // 1,1 et 3,9  
    
    ///////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////En Avant///////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
   
    //Si le joystick est en avant et qu'un des 2 capteurs détecte quelque chose à moins d'un mètre en avant
    //Alors, selon le capteur qui détecte l'obstacle le plus proche on calcule la valeur du signal PWM à créer et on le transmet s'il permet de faire freiner le véhicule
    //Sinon on recopie la valeur lue en analogique.
    if((AN_AA_Val_Volts>2.5) && ((SensorOne_Value<Av_Slow) || (SensorTwo_Value<Av_Slow)))
    {
        distance("Sensor 1 : Avant Gauche ", SensorOne_Value);
        distance("Sensor 2 : Avant Droite ", SensorTwo_Value);
        Serial.println("Obstacle Avant detecte ");
        Serial.print("Valeur Analogique AA lue en octets : ");
        Serial.println(AN_AA_Val_Octet);
        Serial.print("Valeur Analogique convertie en Volts : ");
        Serial.println(AN_AA_Val_Volts);
        if(SensorOne_Value < SensorTwo_Value)
        {
            Serial.println("Obstacle Avant Gauche ");
            PWM_AA_Val_Volts=0.025*SensorOne_Value+1.375;
            if(PWM_AA_Val_Volts <2.5) PWM_AA_Val_Volts=2.5;
            Serial.print("Valeur PWM calculee en Volts : ");
            Serial.println(PWM_AA_Val_Volts);
            if(PWM_AA_Val_Volts<AN_AA_Val_Volts) // Si la valeur calculée est inférieure à la valeur actuelle (inclinaison du joystick)
            {
                Serial.print("On decelere !  Valeur PWM en octets : ");
                PWM_AA_Val_Octet=ConvertVoltsToPWM(PWM_AA_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
                Serial.println(PWM_AA_Val_Octet);
                analogWrite(PWM_AA,PWM_AA_Val_Octet);
            }
            else
            {
                Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : ");              
                PWM_AA_Val_Octet=ConvertVoltsToPWM(AN_AA_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
                Serial.println(PWM_AA_Val_Octet);
                analogWrite(PWM_AA,PWM_AA_Val_Octet);
            }
            
        }
        else if(SensorOne_Value > SensorTwo_Value)
        {
            Serial.println("Obstacle Avant Droit ");
            PWM_AA_Val_Volts=0.025*SensorTwo_Value+1.375;
            if(PWM_AA_Val_Volts <2.5) PWM_AA_Val_Volts=2.5;
            Serial.print("Valeur PWM calculee en Volts : ");
            Serial.println(PWM_AA_Val_Volts);
            if(PWM_AA_Val_Volts<AN_AA_Val_Volts) // Si la valeur calculée est inférieure à la valeur actuelle (inclinaison du joystick)
            {
                Serial.print("On decelere !  Valeur PWM en octets : ");
                PWM_AA_Val_Octet=ConvertVoltsToPWM(PWM_AA_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
                Serial.println(PWM_AA_Val_Octet);
                analogWrite(PWM_AA,PWM_AA_Val_Octet);
            }
            else
            {
                Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : "); 
                PWM_AA_Val_Octet=ConvertVoltsToPWM(AN_AA_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
                Serial.println(PWM_AA_Val_Octet);
                analogWrite(PWM_AA,PWM_AA_Val_Octet);
            }
        } 
    }
    
    
    ///////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////En Arrière/////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    
    
    //Si je vais en arrière (joystick vers l'arrière) et que mon capteur arrière détecte un objet à moins d'un mètre
    //Même stratégie que la marche avant. On recopie la valeur PWM calculée uniquement si elle permet de freiner la marche arrière, sinon on recopie la valeur analogique (forcément plus faible).
    else if((AN_AA_Val_Volts<2.5) && (SensorFour_Value<Ar_Slow))
    {
        distance("Sensor 4 : Arriere ", SensorFour_Value);
        Serial.println("Obstacle Arriere detecte ");
        Serial.print("Valeur Analogique AA lue en octets : ");
        Serial.println(AN_AA_Val_Octet);
        Serial.print("Valeur Analogique convertie en Volts : ");
        Serial.println(AN_AA_Val_Volts);
        Serial.println("Obstacle Arriere ");
        PWM_AA_Val_Volts=-0.025*SensorFour_Value+3.625;
        if(PWM_AA_Val_Volts >2.5) PWM_AA_Val_Volts=2.5;
        Serial.print("Valeur PWM calculee en Volts : ");
        Serial.println(PWM_AA_Val_Volts);
        if(PWM_AA_Val_Volts>AN_AA_Val_Volts) // Si la valeur calculée est supérieure à la valeur actuelle (inclinaison du joystick)
        {
            Serial.print("On decelere !  Valeur PWM en octets : ");
            PWM_AA_Val_Octet=ConvertVoltsToPWM(PWM_AA_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
            Serial.println(PWM_AA_Val_Octet);
            analogWrite(PWM_AA,PWM_AA_Val_Octet);
        }
        else
        {
            Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : ");              
            PWM_AA_Val_Octet=ConvertVoltsToPWM(AN_AA_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
            Serial.println(PWM_AA_Val_Octet);
            analogWrite(PWM_AA,PWM_AA_Val_Octet);
        }
        
    }
    
    
    ///////////////////////////////////////////////////////////////////////////////////////
    //////////////////////Aucun obstacle détecté Avant Arrière/////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    
    //Si on ne détecte aucun obstacle en avant et en arrière et/ou on ne se déplace pas : on recopie la valeur analogique dans le PWM
    else
    {
        PWM_AA_Val_Octet=ConvertVoltsToPWM(AN_AA_Val_Volts);
        PWM_AA_Val_Volts=ConvertPWMToVolts(PWM_AA_Val_Octet);
        analogWrite(PWM_AA,PWM_AA_Val_Octet);
        Serial.print("Analog Volts AA ");
        Serial.println(AN_AA_Val_Volts);
        Serial.print("PWM Volts AA ");
        Serial.println(PWM_AA_Val_Volts);
    }

    ///////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////A Droite///////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    
    //Si le joystick est vers la droite et qu'un des capteurs détecte un obstacle sur la droite
    //Alors on calcule le signal PWM à écrire, s'il permet de freiner le déplacement vers la droite alors on l'écrit.
    //Sinon on recopie le signal analogique d'entrée.
    if((AN_DG_Val_Volts>2.5) && ((SensorThree_Value<Co_Slow) || (SensorFive_Value<Co_Slow)))
    {
        distance("Sensor 3 : Droite Avant ", SensorThree_Value);
        distance("Sensor 5 : Droite Arriere ", SensorFive_Value);
        Serial.println("Obstacle Droite detecte ");
        Serial.print("Valeur Analogique AA lue en octets : ");
        Serial.println(AN_DG_Val_Octet);
        Serial.print("Valeur Analogique convertie en Volts : ");
        Serial.println(AN_DG_Val_Volts);
        if(SensorThree_Value < SensorFive_Value)
        {
            Serial.println("Obstacle Droite Avant ");
            PWM_DG_Val_Volts=0.07*SensorThree_Value+1.45;
            if(PWM_DG_Val_Volts <2.5) PWM_DG_Val_Volts=2.5;
            Serial.print("Valeur PWM calculee en Volts : ");
            Serial.println(PWM_DG_Val_Volts);
            if(PWM_DG_Val_Volts<AN_DG_Val_Volts) // Si la valeur calculée est inférieure à la valeur actuelle (inclinaison du joystick)
            {
                Serial.print("On decelere !  Valeur PWM en octets : ");
                PWM_DG_Val_Octet=ConvertVoltsToPWM(PWM_DG_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }
            else
            {
                Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : ");              
                PWM_DG_Val_Octet=ConvertVoltsToPWM(AN_DG_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }    
        }
        else if(SensorThree_Value > SensorFive_Value)
        {
            Serial.println("Obstacle Droite Arriere ");
            PWM_DG_Val_Volts=0.07*SensorFive_Value+1.45;
            if(PWM_DG_Val_Volts <2.5) PWM_DG_Val_Volts=2.5;
            Serial.print("Valeur PWM calculee en Volts : ");
            Serial.println(PWM_DG_Val_Volts);
            if(PWM_DG_Val_Volts<AN_DG_Val_Volts) // Si la valeur calculée est inférieure à la valeur actuelle (inclinaison du joystick)
            {
                Serial.print("On decelere !  Valeur PWM en octets : ");
                PWM_DG_Val_Octet=ConvertVoltsToPWM(PWM_DG_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }
            else
            {
                Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : "); 
                PWM_DG_Val_Octet=ConvertVoltsToPWM(AN_DG_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }
        } 
    }
    
    
    ///////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////A Gauche///////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    
    //Si le joystick est vers la gauche et qu'un des capteur détecte un obstacle à gauche,
    //Alors on calcule le signal PWM à écrire, s'il permet de faire freiner le mouvement vers la gauche, alors on l'applique
    //Sinon on recopie l'entrée analogique.
    else if((AN_DG_Val_Volts<2.5) && ((SensorSix_Value<Co_Slow) || (SensorSeven_Value<Co_Slow)))
    {
        distance("Sensor 6 : Gauche Arriere ", SensorSix_Value);
        distance("Sensor 7 : Gauche Avant ", SensorSeven_Value);
        Serial.println("Obstacle Gauche detecte ");
        Serial.print("Valeur Analogique AA lue en octets : ");
        Serial.println(AN_DG_Val_Octet);
        Serial.print("Valeur Analogique convertie en Volts : ");
        Serial.println(AN_DG_Val_Volts);
        if(SensorSix_Value < SensorSeven_Value)
        {
            Serial.println("Obstacle Gauche Arriere ");
            PWM_DG_Val_Volts=-0.07*SensorSix_Value+3.55;
            if(PWM_DG_Val_Volts >2.5) PWM_DG_Val_Volts=2.5;
            Serial.print("Valeur PWM calculee en Volts : ");
            Serial.println(PWM_DG_Val_Volts);
            if(PWM_DG_Val_Volts>AN_DG_Val_Volts) // Si la valeur calculée est supérieure à la valeur actuelle (moins d'inclinaison du joystick)
            {
                Serial.print("On decelere !  Valeur PWM en octets : ");
                PWM_DG_Val_Octet=ConvertVoltsToPWM(PWM_DG_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }
            else
            {
                Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : ");              
                PWM_DG_Val_Octet=ConvertVoltsToPWM(AN_DG_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }  
        }
        else if(SensorSix_Value > SensorSeven_Value)
        {
            Serial.println("Obstacle Gauche Avant ");
            PWM_DG_Val_Volts=-0.07*SensorSeven_Value+3.55;
            if(PWM_DG_Val_Volts >2.5) PWM_DG_Val_Volts=2.5;
            Serial.print("Valeur PWM calculee en Volts : ");
            Serial.println(PWM_DG_Val_Volts);
            if(PWM_DG_Val_Volts>AN_DG_Val_Volts) // Si la valeur calculée est supérieure à la valeur actuelle (moins d'inclinaison du joystick)
            {
                Serial.print("On decelere !  Valeur PWM en octets : ");
                PWM_DG_Val_Octet=ConvertVoltsToPWM(PWM_DG_Val_Volts); // On convertit et on envoie la valeur atténuée calculée.
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }
            else
            {
                Serial.print("On ne decelere pas, pas besoin !  Valeur PWM en octets : "); 
                PWM_DG_Val_Octet=ConvertVoltsToPWM(AN_DG_Val_Volts); // On renvoie la valeur actuelle : l'utilisateur va déjà moins vite que la valeur de décélération voulue
                Serial.println(PWM_DG_Val_Octet);
                analogWrite(PWM_DG,PWM_DG_Val_Octet);
            }
        } 
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////
    //////////////////////Aucun obstacle détecté Gauche Droite/////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    
    //Si on ne détecte aucun objet sur les côtés et/ou on ne se déplace pas vers les côtés
    //Alors on recopie l'entrée analogique lue via un signal PWM.
    else
    {
        PWM_DG_Val_Octet=ConvertVoltsToPWM(AN_DG_Val_Volts);
        PWM_DG_Val_Volts=ConvertPWMToVolts(PWM_DG_Val_Octet);
        analogWrite(PWM_DG,PWM_DG_Val_Octet);
        Serial.print("Analog Volts DG ");
        Serial.println(AN_DG_Val_Volts);
        Serial.print("PWM Volts DG ");
        Serial.println(PWM_DG_Val_Volts);
    }
        
    /* Affichage port série
    distance("Sensor 1 : Avant Gauche ", SensorOne_Value);
    distance("Sensor 2 : Avant Droite ", SensorTwo_Value);
    distance("Sensor 3 : Droit Avant ", SensorThree_Value);
    distance("Sensor 4 : Droit Arriere", SensorFour_Value);
    distance("Sensor 5 : Arriere ", SensorFive_Value);
    distance("Sensor 6 : Gauche Arriere ", SensorSix_Value);
    distance("Sensor 7 : Gauche Avant ", SensorSeven_Value);
    */
}



//////////////////////////////////////////////////////////////////////
//////////////////FONCTIONS ANNEXES///////////////////////////////////
//////////////////////////////////////////////////////////////////////

//Permet de convertir une donnée analogique lue sur un port A en Volts (déterminé par l'expérimentation).
float ConvertAnalogToVolts(int Analog)
{
    return Analog*3.44/679;
}

//Permet de convertir une donnée en volts vers un rapport cyclique pour un signal PWM (déterminé par l'expérimentation)
int ConvertVoltsToPWM(float Volts)
{
    return Volts*52/1.1;
}

//Permet de convertir une donnée en rapport cyclique PWM vers des volts (déterminé par l'expérimentation)
float ConvertPWMToVolts(int PWMOctet)
{
     return PWMOctet*1.1/52;
}

//Permet de configurer un capteur SRF08 pour régler sa portée à 2 mètres et son gain à 0.
void setupSensor(SonarSRF08 Sensor, int Range)
{
    Sensor.setRange(Range);
    Sensor.setGain(0x01);
    delay(250);
    Sensor.setRange(Range);
    Sensor.setGain(0x00);
  
}

//Affichage de la distance d'un capteur.
void distance(String reference, int sensorReading) {
    Serial.print("Distance from " + reference + ": ");
    Serial.print(sensorReading);
    Serial.println(unit);
}

//Test de connexion d'un capteur avec récupération du soft
void isConnected(String reference, int sensorSoft) {
    if (sensorSoft >= 0)
    {
        Serial.print("Sensor " + reference + " connected (");
        Serial.print(sensorSoft);
        Serial.println(")");
    }
    else
    {
        Serial.println("Sensor " + reference + " not detected");
    }
}


/**
 * Divides a given PWM pin frequency by a divisor.
 * 
 * The resulting frequency is equal to the base frequency divided by
 * the given divisor:
 *   - Base frequencies:
 *      o The base frequency for pins 3, 9, 10, and 11 is 31250 Hz.
 *      o The base frequency for pins 5 and 6 is 62500 Hz.
 *   - Divisors:
 *      o The divisors available on pins 5, 6, 9 and 10 are: 1, 8, 64,
 *        256, and 1024.
 *      o The divisors available on pins 3 and 11 are: 1, 8, 32, 64,
 *        128, 256, and 1024.
 * 
 * PWM frequencies are tied together in pairs of pins. If one in a
 * pair is changed, the other is also changed to match:
 *   - Pins 5 and 6 are paired on timer0
 *   - Pins 9 and 10 are paired on timer1
 *   - Pins 3 and 11 are paired on timer2
 * 
 * Note that this function will have side effects on anything else
 * that uses timers:
 *   - Changes on pins 3, 5, 6, or 11 may cause the delay() and
 *     millis() functions to stop working. Other timing-related
 *     functions may also be affected.
 *   - Changes on pins 9 or 10 will cause the Servo library to function
 *     incorrectly.
 * 
 * Thanks to macegr of the Arduino forums for his documentation of the
 * PWM frequency divisors. His post can be viewed at:
 *   http://www.arduino.cc/cgi-bin/yabb2/YaBB.pl?num=1235060559/0#4
 */
void setPwmFrequency(int pin, int divisor) {
  byte mode;
  if(pin == 5 || pin == 6 || pin == 9 || pin == 10) {
    switch(divisor) {
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 64: mode = 0x03; break;
      case 256: mode = 0x04; break;
      case 1024: mode = 0x05; break;
      default: return;
    }
    if(pin == 5 || pin == 6) {
      TCCR0B = TCCR0B & 0b11111000 | mode;
    } else {
      TCCR1B = TCCR1B & 0b11111000 | mode;
    }
  } else if(pin == 3 || pin == 11) {
    switch(divisor) {
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 32: mode = 0x03; break;
      case 64: mode = 0x04; break;
      case 128: mode = 0x05; break;
      case 256: mode = 0x06; break;
      case 1024: mode = 0x7; break;
      default: return;
    }
    TCCR2B = TCCR2B & 0b11111000 | mode;
  }
}
